"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getInstructionFunctionHighLevelFragment = void 0;
const shared_1 = require("../../../shared");
const ContextMap_1 = require("../ContextMap");
const common_1 = require("./common");
const instructionBytesCreatedOnChain_1 = require("./instructionBytesCreatedOnChain");
const instructionInputResolved_1 = require("./instructionInputResolved");
const instructionRemainingAccounts_1 = require("./instructionRemainingAccounts");
function getInstructionFunctionHighLevelFragment(instructionNode, programNode, renamedArgs, dataArgsManifest, resolvedInputs) {
    const hasAccounts = instructionNode.accounts.length > 0;
    const hasDataArgs = !!instructionNode.dataArgs.link ||
        instructionNode.dataArgs.struct.fields.filter((field) => { var _a; return ((_a = field.defaultsTo) === null || _a === void 0 ? void 0 : _a.strategy) !== 'omitted'; }).length > 0;
    const hasExtraArgs = !!instructionNode.extraArgs.link ||
        instructionNode.extraArgs.struct.fields.filter((field) => { var _a; return ((_a = field.defaultsTo) === null || _a === void 0 ? void 0 : _a.strategy) !== 'omitted'; }).length > 0;
    const hasAnyArgs = hasDataArgs || hasExtraArgs;
    const argsTypeFragment = (0, common_1.fragment)(instructionNode.dataArgs.link
        ? dataArgsManifest.looseType.render
        : `${(0, shared_1.pascalCase)(instructionNode.dataArgs.name)}Args`);
    if (instructionNode.dataArgs.link) {
        argsTypeFragment.mergeImportsWith(dataArgsManifest.looseType);
    }
    const functionName = (0, shared_1.camelCase)(instructionNode.name);
    const lowLevelFunctionName = `${functionName}Instruction`;
    const typeParamsFragment = getTypeParams(instructionNode, programNode);
    const instructionTypeFragment = getInstructionType(instructionNode);
    const inputTypeFragment = getInputType(instructionNode);
    const customGeneratedInstruction = `CustomGeneratedInstruction<${instructionTypeFragment.render}, TReturn>`;
    const renamedArgsText = [...renamedArgs.entries()]
        .map(([k, v]) => `${k}: input.${v}`)
        .join(', ');
    const resolvedInputsFragment = (0, instructionInputResolved_1.getInstructionInputResolvedFragment)(instructionNode, resolvedInputs);
    const remainingAccountsFragment = (0, instructionRemainingAccounts_1.getInstructionRemainingAccountsFragment)(instructionNode);
    const bytesCreatedOnChainFragment = (0, instructionBytesCreatedOnChain_1.getInstructionBytesCreatedOnChainFragment)(instructionNode);
    const context = new ContextMap_1.ContextMap()
        .add('getProgramAddress')
        .mergeWith(resolvedInputsFragment.interfaces, remainingAccountsFragment.interfaces, bytesCreatedOnChainFragment.interfaces);
    const contextFragment = context.toFragment();
    const functionFragment = (0, common_1.fragmentFromTemplate)('instructionFunctionHighLevel.njk', {
        instruction: instructionNode,
        program: programNode,
        hasAccounts,
        hasDataArgs,
        hasExtraArgs,
        hasAnyArgs,
        argsType: argsTypeFragment,
        functionName,
        lowLevelFunctionName,
        typeParams: typeParamsFragment,
        instructionType: instructionTypeFragment,
        inputType: inputTypeFragment,
        context: contextFragment,
        renamedArgs: renamedArgsText,
        resolvedInputs: resolvedInputsFragment,
        remainingAccounts: remainingAccountsFragment,
        bytesCreatedOnChain: bytesCreatedOnChainFragment,
        customGeneratedInstruction,
    })
        .mergeImportsWith(typeParamsFragment, instructionTypeFragment, inputTypeFragment, contextFragment, resolvedInputsFragment, remainingAccountsFragment, bytesCreatedOnChainFragment, argsTypeFragment)
        .addImports('solanaAddresses', ['Base58EncodedAddress'])
        .addImports('shared', ['CustomGeneratedInstruction', 'WrappedInstruction']);
    if (hasAccounts) {
        functionFragment
            .addImports('solanaInstructions', ['IAccountMeta'])
            .addImports('shared', ['getAccountMetasAndSigners', 'ResolvedAccount']);
    }
    return functionFragment;
}
exports.getInstructionFunctionHighLevelFragment = getInstructionFunctionHighLevelFragment;
function getTypeParams(instructionNode, programNode) {
    const typeParams = [
        ...instructionNode.accounts.map((account) => `TAccount${(0, shared_1.pascalCase)(account.name)} extends string`),
        `TProgram extends string = "${programNode.publicKey}"`,
    ];
    return (0, common_1.fragment)(typeParams.filter((x) => !!x).join(', '));
}
function getInstructionType(instructionNode) {
    const instructionTypeName = (0, shared_1.pascalCase)(`${instructionNode.name}Instruction`);
    const accountTypeParamsFragments = instructionNode.accounts.map((account) => {
        const typeParam = `TAccount${(0, shared_1.pascalCase)(account.name)}`;
        const camelName = (0, shared_1.camelCase)(account.name);
        if (account.isSigner === 'either') {
            const role = account.isWritable
                ? 'WritableSignerAccount'
                : 'ReadonlySignerAccount';
            return (0, common_1.fragment)(`typeof input["${camelName}"] extends Signer<${typeParam}> ? ${role}<${typeParam}> : ${typeParam}`).addImports('solanaInstructions', [role]);
        }
        return (0, common_1.fragment)(typeParam);
    });
    return (0, common_1.mergeFragments)([(0, common_1.fragment)('TProgram'), ...accountTypeParamsFragments], (renders) => renders.join(', ')).mapRender((r) => `${instructionTypeName}<${r}>`);
}
function getInputType(instructionNode) {
    const inputTypeName = (0, shared_1.pascalCase)(`${instructionNode.name}Input`);
    if (instructionNode.accounts.length === 0)
        return (0, common_1.fragment)(inputTypeName);
    const accountTypeParams = instructionNode.accounts
        .map((account) => `TAccount${(0, shared_1.pascalCase)(account.name)}`)
        .join(', ');
    return (0, common_1.fragment)(`${inputTypeName}<${accountTypeParams}>`);
}
//# sourceMappingURL=instructionFunctionHighLevel.js.map