"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.renderJavaScriptInstructionDefaults = void 0;
const shared_1 = require("../../shared");
const JavaScriptContextMap_1 = require("./JavaScriptContextMap");
const JavaScriptImportMap_1 = require("./JavaScriptImportMap");
const RenderJavaScriptValueNode_1 = require("./RenderJavaScriptValueNode");
function renderJavaScriptInstructionDefaults(input, optionalAccountStrategy, argObject) {
    const imports = new JavaScriptImportMap_1.JavaScriptImportMap();
    const interfaces = new JavaScriptContextMap_1.JavaScriptContextMap();
    if (!input.defaultsTo) {
        return { imports, interfaces, render: '' };
    }
    const { defaultsTo } = input;
    const render = (defaultValue, isWritable) => {
        const inputName = (0, shared_1.camelCase)(input.name);
        if (input.kind === 'account' && defaultsTo.kind === 'resolver') {
            return {
                imports,
                interfaces,
                render: `resolvedAccounts.${inputName} = { ...resolvedAccounts.${inputName}, ...${defaultValue} };`,
            };
        }
        if (input.kind === 'account' && isWritable === undefined) {
            return {
                imports,
                interfaces,
                render: `resolvedAccounts.${inputName}.value = ${defaultValue};`,
            };
        }
        if (input.kind === 'account') {
            return {
                imports,
                interfaces,
                render: `resolvedAccounts.${inputName}.value = ${defaultValue};\n` +
                    `resolvedAccounts.${inputName}.isWritable = ${isWritable ? 'true' : 'false'}`,
            };
        }
        return {
            imports,
            interfaces,
            render: `${argObject}.${inputName} = ${defaultValue};`,
        };
    };
    switch (defaultsTo.kind) {
        case 'account':
            const name = (0, shared_1.camelCase)(defaultsTo.name);
            if (input.kind === 'account') {
                imports.add('shared', 'expectSome');
                if (input.resolvedIsSigner && !input.isSigner) {
                    return render(`expectSome(resolvedAccounts.${name}.value).publicKey`);
                }
                return render(`expectSome(resolvedAccounts.${name}.value)`);
            }
            imports.add('shared', 'expectPublicKey');
            return render(`expectPublicKey(resolvedAccounts.${name}.value)`);
        case 'pda':
            const pdaFunction = `find${(0, shared_1.pascalCase)(defaultsTo.pdaAccount)}Pda`;
            const pdaImportFrom = defaultsTo.importFrom === 'generated'
                ? 'generatedAccounts'
                : defaultsTo.importFrom;
            imports.add(pdaImportFrom, pdaFunction);
            interfaces.add('eddsa');
            const pdaArgs = ['context'];
            const pdaSeeds = Object.keys(defaultsTo.seeds).map((seed) => {
                const seedValue = defaultsTo.seeds[seed];
                if (seedValue.kind === 'account') {
                    imports.add('shared', 'expectPublicKey');
                    return `${seed}: expectPublicKey(resolvedAccounts.${(0, shared_1.camelCase)(seedValue.name)}.value)`;
                }
                if (seedValue.kind === 'arg') {
                    imports.add('shared', 'expectSome');
                    return `${seed}: expectSome(${argObject}.${(0, shared_1.camelCase)(seedValue.name)})`;
                }
                const valueManifest = (0, RenderJavaScriptValueNode_1.renderJavaScriptValueNode)(seedValue.value);
                imports.mergeWith(valueManifest.imports);
                return `${seed}: ${valueManifest.render}`;
            });
            if (pdaSeeds.length > 0) {
                pdaArgs.push(`{ ${pdaSeeds.join(', ')} }`);
            }
            return render(`${pdaFunction}(${pdaArgs.join(', ')})`);
        case 'publicKey':
            imports.add('umi', 'publicKey');
            return render(`publicKey('${defaultsTo.publicKey}')`);
        case 'program':
            return render(`context.programs.getPublicKey('${defaultsTo.program.name}', '${defaultsTo.program.publicKey}')`, false);
        case 'programId':
            if (optionalAccountStrategy === 'programId' &&
                input.kind === 'account' &&
                input.isOptional) {
                return { imports, interfaces, render: '' };
            }
            return render('programId', false);
        case 'identity':
            interfaces.add('identity');
            if (input.kind === 'account' && input.isSigner !== false) {
                return render('context.identity');
            }
            return render('context.identity.publicKey');
        case 'payer':
            interfaces.add('payer');
            if (input.kind === 'account' && input.isSigner !== false) {
                return render('context.payer');
            }
            return render('context.payer.publicKey');
        case 'accountBump':
            imports.add('shared', 'expectPda');
            return render(`expectPda(resolvedAccounts.${(0, shared_1.camelCase)(defaultsTo.name)}.value)[1]`);
        case 'arg':
            imports.add('shared', 'expectSome');
            return render(`expectSome(${argObject}.${(0, shared_1.camelCase)(defaultsTo.name)})`);
        case 'value':
            const valueManifest = (0, RenderJavaScriptValueNode_1.renderJavaScriptValueNode)(defaultsTo.value);
            imports.mergeWith(valueManifest.imports);
            return render(valueManifest.render);
        case 'resolver':
            const resolverName = (0, shared_1.camelCase)(defaultsTo.name);
            const isWritable = input.kind === 'account' && input.isWritable ? 'true' : 'false';
            imports.add(defaultsTo.importFrom, resolverName);
            interfaces.add(['eddsa', 'identity', 'payer']);
            return render(`${resolverName}(context, resolvedAccounts, ${argObject}, programId, ${isWritable})`);
        case 'conditional':
        case 'conditionalResolver':
            const ifTrueRenderer = renderNestedInstructionDefault(input, optionalAccountStrategy, defaultsTo.ifTrue, argObject);
            const ifFalseRenderer = renderNestedInstructionDefault(input, optionalAccountStrategy, defaultsTo.ifFalse, argObject);
            if (!ifTrueRenderer && !ifFalseRenderer) {
                return { imports, interfaces, render: '' };
            }
            if (ifTrueRenderer) {
                imports.mergeWith(ifTrueRenderer.imports);
                interfaces.mergeWith(ifTrueRenderer.interfaces);
            }
            if (ifFalseRenderer) {
                imports.mergeWith(ifFalseRenderer.imports);
                interfaces.mergeWith(ifFalseRenderer.interfaces);
            }
            const negatedCondition = !ifTrueRenderer;
            let condition = 'true';
            if (defaultsTo.kind === 'conditional') {
                const comparedInputName = defaultsTo.input.kind === 'account'
                    ? `resolvedAccounts.${(0, shared_1.camelCase)(defaultsTo.input.name)}.value`
                    : `${argObject}.${(0, shared_1.camelCase)(defaultsTo.input.name)}`;
                if (defaultsTo.value) {
                    const comparedValue = (0, RenderJavaScriptValueNode_1.renderJavaScriptValueNode)(defaultsTo.value);
                    imports.mergeWith(comparedValue.imports);
                    const operator = negatedCondition ? '!==' : '===';
                    condition = `${comparedInputName} ${operator} ${comparedValue.render}`;
                }
                else {
                    condition = negatedCondition
                        ? `!${comparedInputName}`
                        : comparedInputName;
                }
            }
            else {
                const conditionalResolverName = (0, shared_1.camelCase)(defaultsTo.resolver.name);
                const conditionalIsWritable = input.kind === 'account' && input.isWritable ? 'true' : 'false';
                imports.add(defaultsTo.resolver.importFrom, conditionalResolverName);
                interfaces.add(['eddsa', 'identity', 'payer']);
                condition = `${conditionalResolverName}(context, resolvedAccounts, ${argObject}, programId, ${conditionalIsWritable})`;
                condition = negatedCondition ? `!${condition}` : condition;
            }
            if (ifTrueRenderer && ifFalseRenderer) {
                return {
                    imports,
                    interfaces,
                    render: `if (${condition}) {\n${ifTrueRenderer.render}\n} else {\n${ifFalseRenderer.render}\n}`,
                };
            }
            return {
                imports,
                interfaces,
                render: `if (${condition}) {\n${ifTrueRenderer ? ifTrueRenderer.render : ifFalseRenderer === null || ifFalseRenderer === void 0 ? void 0 : ifFalseRenderer.render}\n}`,
            };
        default:
            const neverDefault = defaultsTo;
            throw new Error(`Unexpected value type ${neverDefault.kind}`);
    }
}
exports.renderJavaScriptInstructionDefaults = renderJavaScriptInstructionDefaults;
function renderNestedInstructionDefault(input, optionalAccountStrategy, defaultsTo, argObject) {
    if (!defaultsTo)
        return undefined;
    if (input.kind === 'account') {
        return renderJavaScriptInstructionDefaults({ ...input, defaultsTo: defaultsTo }, optionalAccountStrategy, argObject);
    }
    return renderJavaScriptInstructionDefaults({ ...input, defaultsTo: defaultsTo }, optionalAccountStrategy, argObject);
}
//# sourceMappingURL=RenderJavaScriptInstructionDefaults.js.map